import json
import logging

from nameko_tracer import constants


def default(obj):
    return str(obj)


def serialise(obj):
    return json.dumps(obj, default=default)


class JSONFormatter(logging.Formatter):
    """ Format trace data as JSON string
    """

    def format(self, record):
        return serialise(getattr(record, constants.TRACE_KEY))


class ElasticsearchDocumentFormatter(JSONFormatter):
    """ Format trace as JSON which can be fed to Elasticsearch as a document

    Request and response data fields of the document are serialized as JSON
    string before serialising the whole output.

    """

    extra_serialise_keys = (
        constants.CONTEXT_DATA_KEY,
        constants.REQUEST_KEY,
        constants.RESPONSE_KEY,
        constants.EXCEPTION_ARGS_KEY)

    def format(self, record):

        trace = getattr(record, constants.TRACE_KEY)

        for key in self.extra_serialise_keys:
            if key in trace:
                trace[key] = serialise(trace[key])

        return serialise(trace)



import abc
import logging
import re


from nameko_tracer import constants, utils


class BaseTruncateFilter(logging.Filter, abc.ABC):

    default_entrypoints = []

    def __init__(self, entrypoints=None, max_len=None):

        entrypoints = entrypoints or self.default_entrypoints
        self.entrypoints = [re.compile(r) for r in entrypoints]

        self.max_len = max_len or 100

    def filter(self, log_record):
        data = getattr(log_record, constants.TRACE_KEY)
        entrypoint_name = data.get(constants.ENTRYPOINT_NAME_KEY)
        if any(regex.match(entrypoint_name) for regex in self.entrypoints):
            data = self.truncate(data)
            setattr(log_record, constants.TRACE_KEY, data)
        return log_record

    @abc.abstractmethod
    def truncate(self, data):
        """ Truncate and return the data
        """


class TruncateCallArgsFilter(BaseTruncateFilter):
    """ Truncate serialized call arguments

    If the truncation is applied, the call data is serialised to string
    beforehand.

    Example of a filter truncating call arguments of entrypoint methods
    starting with "create" or "insert" to the length of 200 characters::

        filter = TruncateRequestFilter(
            entrypoints=['^create|^insert'], max_len=200)

    """

    default_entrypoints = []

    def truncate(self, data):
        if constants.REQUEST_KEY not in data:
            return data

        call_args = utils.serialise_to_string(data[constants.REQUEST_KEY])
        length = len(call_args)
        if length > self.max_len:
            data[constants.REQUEST_KEY] = call_args[:self.max_len]
            truncated = True
        else:
            truncated = False
        data[constants.REQUEST_TRUNCATED_KEY] = truncated
        data[constants.REQUEST_LENGTH_KEY] = length
        return data


class TruncateResponseFilter(BaseTruncateFilter):
    """ Truncate serialized response data

    If the truncation is applied, the call data is serialised to string
    beforehand.

    Example of a filter truncating return value of entrypoint methods
    starting with "get" or "list" to the length of 200 characters::

        filter = TruncateResponseFilter(
            entrypoints=['^get|^list'], max_len=200)

    """

    default_entrypoints = ['^get_|^list_|^query_']

    def truncate(self, data):

        if constants.RESPONSE_KEY not in data:
            return data

        result = utils.serialise_to_string(data[constants.RESPONSE_KEY])
        length = len(result)
        if length > self.max_len:
            data[constants.RESPONSE_KEY] = result[:self.max_len]
            truncated = True
        else:
            truncated = False
        data[constants.RESPONSE_TRUNCATED_KEY] = truncated
        data[constants.RESPONSE_LENGTH_KEY] = length
        return data


TruncateRequestFilter = TruncateCallArgsFilter



from enum import Enum


class Stage(Enum):
    """ Entrypoint stage
    """
    request = 'request'
    response = 'response'


class Status(Enum):
    """ Entrypoint result status
    """
    success = 'success'
    error = 'error'


LOGGER_NAME = 'nameko_tracer'
""" Name of the logger used for entrypoint logging

Use this name to configure entrypoint logging in ``LOGGING`` setting
of Nameko config.
"""


TRACE_KEY = 'nameko_trace'
""" Name of the log record attribute holding the serialisable details

Contains gathered entrypoint call and result details in serialisable
and compact form.
"""


TIMESTAMP_KEY = 'timestamp'
""" A key holding the entrypoint stage timestamp
"""


STAGE_KEY = 'stage'
""" A key holding the lifecycle stage (a value of one of ``Stage`` options)
"""


HOSTNAME_KEY = 'hostname'
""" A key holding the service host name
"""


REQUEST_KEY = 'call_args'
""" A key holding a dictionary of arguments passed to the entrypoint call
"""


REQUEST_REDACTED_KEY = 'call_args_redacted'
"""
A key holding a boolean value saying whether sensitive values of the
entrypoint call arguments were redacted.
"""


RESPONSE_KEY = 'response'
""" A key holding serialisable return value of the entrypoint.
"""


REQUEST_TRUNCATED_KEY = 'call_args_truncated'
"""
A key holding a boolean value saying whether the call args data were
truncated. Set by ``TruncateRequestFilter``.
"""


RESPONSE_TRUNCATED_KEY = 'response_truncated'
"""
A key holding a boolean value saying whether the result data were truncated.
Set by ``TruncateResponseFilter``.
"""


REQUEST_LENGTH_KEY = 'call_args_length'
""" A key holding the original call args data length

Set by ``TruncateRequestFilter`` to the original length of data in
``REQUEST_KEY``.
"""

RESPONSE_LENGTH_KEY = 'response_length'
""" A key holding the original result data length

Set by ``TruncateResponseFilter`` to the original length of data in
``RESPONSE_KEY``.
"""


RESPONSE_STATUS_KEY = 'response_status'
""" A key holding the result status (a value of one of ``Status`` options)
"""


RESPONSE_TIME_KEY = 'response_time'
""" A key holding the amount of time taken between the two stages
"""


EXCEPTION_TYPE_KEY = 'exception_type'
""" A key holding exception type name

Set if the entrypoint resulted into an error
"""


EXCEPTION_PATH_KEY = 'exception_path'
""" A key holding exception path e.g. ``some.module.SomeError``

Set if the entrypoint resulted into an error
"""


EXCEPTION_VALUE_KEY = 'exception_value'
""" A key holding string representation of exception raised

Set if the entrypoint resulted into an error
"""


EXCEPTION_ARGS_KEY = 'exception_args'
""" A key holding a list of exception arguments

Set if the entrypoint resulted into an error
"""


EXCEPTION_TRACEBACK_KEY = 'exception_traceback'
""" A key holding exception traceback string

Set if the entrypoint resulted into an error
"""


EXCEPTION_EXPECTED_KEY = 'exception_expected'
""" A key holding a boolean saying whether the exception raised was one of
errors expected by the entrypoint
"""


SERVICE_NAME_KEY = 'service'
""" A key holding the name of the service
"""


ENTRYPOINT_NAME_KEY = 'entrypoint_name'
""" A key holding the entrypoint service method name e.g. ``'get_user'``
"""


ENTRYPOINT_TYPE_KEY = 'entrypoint_type'
""" A key holding the entrypoint type name e.g. ``'Rpc'``.
"""


CALL_ID_KEY = 'call_id'
""" A key holding the unique ID of the entrypoint call
"""

CALL_ID_STACK_KEY = 'call_id_stack'
""" A key holding the call ID stack ...
"""

ORIGIN_CALL_ID_KEY = 'origin_call_id'
""" A key holding the Source of all stack calls
"""

CONTEXT_DATA_KEY = 'context_data'
""" A key holding the worker context data dictionary
"""


DEFAULT_ADAPTERS = {
    'nameko.web.handlers.HttpRequestHandler': (
        'nameko_tracer.adapters.HttpRequestHandlerAdapter'),
}
"""
Default adapter overrides setup

Sets an override for Nameko built-in HttpRequestHandler. Extra overrides
coming from config are merged in.
"""


CONFIG_KEY = 'TRACER'
""" Nameko config key holding tracer configuration
"""


ADAPTERS_CONFIG_KEY = 'ADAPTERS'
"""
A key holding adapters configuration in Nameko config (under ``CONFIG_KEY``)
"""



from collections import defaultdict
from datetime import datetime
import logging
import socket
from weakref import WeakKeyDictionary

from nameko.extensions import DependencyProvider

from nameko_tracer import adapters, constants, utils


logger = logging.getLogger(__name__)


class Tracer(DependencyProvider):
    """ Entrypoint logging dependency

    Logs call and result details about entrypoints fired.

    """

    def __init__(self):
        self.logger = None
        self.adapter_types = defaultdict(lambda: adapters.DefaultAdapter)
        self.worker_timestamps = WeakKeyDictionary()

    def setup(self):
        config = self.container.config.get(constants.CONFIG_KEY, {})

        self.configure_adapter_types(constants.DEFAULT_ADAPTERS)
        self.configure_adapter_types(
            config.get(constants.ADAPTERS_CONFIG_KEY, {}))

        self.logger = logging.getLogger(constants.LOGGER_NAME)

    def configure_adapter_types(self, adapters_config):
        for entrypoint_path, adapter_path in adapters_config.items():
            entrypoint_class = utils.import_by_path(entrypoint_path)
            adapter_class = utils.import_by_path(adapter_path)
            self.adapter_types[entrypoint_class] = adapter_class

    def adapter_factory(self, worker_ctx):
        adapter_class = self.adapter_types[type(worker_ctx.entrypoint)]
        extra = {'hostname': socket.gethostname()}
        return adapter_class(self.logger, extra=extra)

    def worker_setup(self, worker_ctx):
        """ Log entrypoint call details
        """

        timestamp = datetime.utcnow()
        self.worker_timestamps[worker_ctx] = timestamp

        try:
            extra = {
                'stage': constants.Stage.request,
                'worker_ctx': worker_ctx,
                'timestamp': timestamp,
            }
            adapter = self.adapter_factory(worker_ctx)
            adapter.info(
                '[%s] entrypoint call trace',
                worker_ctx.call_id,
                extra=extra)
        except Exception:
            logger.warning('Failed to log entrypoint trace', exc_info=True)

    def worker_result(self, worker_ctx, result=None, exc_info=None):
        """ Log entrypoint result details
        """

        timestamp = datetime.utcnow()
        worker_setup_timestamp = self.worker_timestamps[worker_ctx]
        response_time = (timestamp - worker_setup_timestamp).total_seconds()

        try:
            extra = {
                'stage': constants.Stage.response,
                'worker_ctx': worker_ctx,
                'result': result,
                'exc_info_': exc_info,
                'timestamp': timestamp,
                'response_time': response_time,
            }
            adapter = self.adapter_factory(worker_ctx)
            if exc_info:
                adapter.warning(
                    '[%s] entrypoint result trace',
                    worker_ctx.call_id,
                    extra=extra)
            else:
                adapter.info(
                    '[%s] entrypoint result trace',
                    worker_ctx.call_id,
                    extra=extra)
        except Exception:
            logger.warning('Failed to log entrypoint trace', exc_info=True)



#!/usr/bin/env python
from setuptools import find_packages, setup

setup(
    name='nameko-tracer',
    version='1.2.0',
    description='Nameko extension logging entrypoint processing metrics',
    author='student.com',
    author_email='wearehiring@student.com',
    url='https://github.com/Overseas-Student-Living/nameko-tracer',
    packages=find_packages(exclude=['test', 'test.*']),
    install_requires=[
        "nameko>=2.2.0",
    ],
    extras_require={
        'dev': [
            "coverage==4.4.1",
            "flake8==3.4.1",
            "pylint==1.7.2",
            "pytest==3.2.0",
        ]
    },
    dependency_links=[],
    zip_safe=True,
    license='Apache License, Version 2.0'
)
